/*
 GBCartRead
 Version: 1.7 Rev 1
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 18/03/2011
 Last Modified: 17/02/2014
 
 GBCartRead is an Arduino based Gameboy Cartridge Reader which uses a C program or python script to interface with 
 the Arduino. GBCartRead allows you to dump your ROM, save the RAM and write to the RAM.

 Works with Arduino Duemilanove and Uno. Will work for other 5V Arduinos but requires wiring changes.

 Speed increase thanks to Frode vdM. (fvdm1992@yahoo.no) and David R
*/

// Edit these in pindelcarations.h too
int latchPin = 10;
int dataPin = 11;
int clockPin = 13;
int rdPin = A5;
int wrPin = A3;
int mreqPin = A4;
int mosfetControlPin = A0;

#include "pindeclarations.h"
#include <SPI.h>
#include <util/delay.h>

char gameTitle[17];
int cartridgeType = 0;
int romSize = 0;
int romBanks = 0;
int ramSize = 0;
int ramBanks = 0;

void setup() {
  pinMode(latchPin, OUTPUT);
  pinMode(clockPin, OUTPUT);
  pinMode(dataPin, OUTPUT);
  pinMode(rdPin, OUTPUT);
  pinMode(wrPin, OUTPUT);
  pinMode(mreqPin, OUTPUT);
  pinMode(mosfetControlPin, OUTPUT);
  PORTC |= (1<<PC0); // Turn mosfet off
  
  // Set pins as inputs
  DDRB &= ~((1<<PB0) | (1<<PB1)); // D8 & D9
  DDRD &= ~((1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7)); // D2 to D7
  rd_wr_mreq_off();
  
  // Setup pin interrupt for button
  PCMSK1 = (1<<PCINT9);
  PCICR = (1<<PCIE1);
  
  Serial.begin(400000);
}

void loop() {
  // Wait for serial input
  while (Serial.available() <= 0) {
    delay(200);
  }

  // Decode input
  char readInput[10];
  int readCount = 0;
  while (Serial.available() > 0) {
    char c = Serial.read();
    readInput[readCount] = c;
    readCount++;
  }
  readInput[readCount] = '\0';
  
  // Cartridge Header
  if (strstr(readInput, "HEADER")) {
    rd_wr_mreq_reset();
    
    // Read Cartridge Header
    for (int addr = 0x0134; addr <= 0x143; addr++) {
      gameTitle[(addr-0x0134)] = (char) readByte(addr);
    }
    gameTitle[16] = '\0';
    cartridgeType = readByte(0x0147);
    romSize = readByte(0x0148);
    ramSize = readByte(0x0149);
    romBanks = 2; // Default 32K
    if (romSize == 1) { romBanks = 4; } 
    if (romSize == 2) { romBanks = 8; } 
    if (romSize == 3) { romBanks = 16; } 
    if (romSize == 4) { romBanks = 32; } 
    if (romSize == 5 && (cartridgeType == 1 || cartridgeType == 2 || cartridgeType == 3)) { romBanks = 63; } 
    else if (romSize == 5) { romBanks = 64; } 
    if (romSize == 6 && (cartridgeType == 1 || cartridgeType == 2 || cartridgeType == 3)) { romBanks = 125; } 
    else if (romSize == 6) { romBanks = 128; }
    if (romSize == 7) { romBanks = 256; }
    if (romSize == 82) { romBanks = 72; }
    if (romSize == 83) { romBanks = 80; }
    if (romSize == 84) { romBanks = 96; }
    ramBanks = 1; // Default 8K RAM
    if (ramSize == 3) { ramBanks = 4; }
    if (ramSize == 4){ ramBanks = 16; } // GB Camera
    
    Serial.println(gameTitle);
    Serial.println(cartridgeType);
    Serial.println(romSize);
    Serial.println(ramSize);
  }

  // Dump ROM
  else if (strstr(readInput, "READROM")) {
    rd_wr_mreq_reset();
    unsigned int addr = 0;
    
    // Read x number of banks
    for (int y = 1; y < romBanks; y++) {
      writeByte(0x2100, y); // Set ROM bank
      if (y > 1) {addr = 0x4000;}
      for (; addr <= 0x7FFF; addr = addr+64) {
        uint8_t readData[64];
        for (int i = 0; i < 64; i++){
          readData[i] = readByte(addr+i);
        }
        Serial.write(readData, 64);
      }
    }
  }
  
  // Read RAM
  else if (strstr(readInput, "READRAM")) {
    rd_wr_mreq_reset();
    
    // MBC2 Fix (unknown why this fixes it, maybe has to read ROM before RAM?)
    readByte(0x0134);

    unsigned int addr = 0;
    unsigned int endaddr = 0;
    if (cartridgeType == 6 && ramSize == 0) { endaddr = 0xA1FF; } // MBC2 512bytes (nibbles)
    if (ramSize == 1) { endaddr = 0xA7FF; } // 2K RAM
    if (ramSize > 1) { endaddr = 0xBFFF; } // 8K RAM
    
    // Does cartridge have RAM
    if (endaddr > 0) {
      // Initialise MBC
      writeByte(0x0000, 0x0A);

      // Switch RAM banks
      for (int bank = 0; bank < ramBanks; bank++) {
        writeByte(0x4000, bank);

        // Read RAM
        for (addr = 0xA000; addr <= endaddr; addr = addr+64) {  
          uint8_t readData[64];
          for (int i = 0; i < 64; i++){
            readData[i] = readByte(addr+i);
          }
          Serial.write(readData, 64);
        }
      }
      
      // Disable RAM
      writeByte(0x0000, 0x00);
    }
  }
  
  // Write RAM
  else if (strstr(readInput, "WRITERAM")) {
    rd_wr_mreq_reset();
    
    // MBC2 Fix (unknown why this fixes it, maybe has to read ROM before RAM?)
    readByte(0x0134);
    unsigned int addr = 0;
    unsigned int endaddr = 0;
    if (cartridgeType == 6 && ramSize == 0) { endaddr = 0xA1FF; } // MBC2 512bytes (nibbles)
    if (ramSize == 1) { endaddr = 0xA7FF; } // 2K RAM
    if (ramSize > 1) { endaddr = 0xBFFF; } // 8K RAM
    
    // Does cartridge have RAM
    if (endaddr > 0) {
      // Initialise MBC
      writeByte(0x0000, 0x0A);
      
      // Switch RAM banks
      for (int bank = 0; bank < ramBanks; bank++) {
        writeByte(0x4000, bank);
        
        // Write RAM
        for (addr = 0xA000; addr <= endaddr; addr=addr+64) {  
          
          // Wait for serial input
          for (uint8_t i = 0; i < 64; i++) {
            // Wait for serial input
            while (Serial.available() <= 0);
            
            // Read input
            uint8_t bval = (uint8_t) Serial.read();
            
            // Write to RAM
            mreqPin_low;
            writeByte(addr+i, bval);
            asm volatile("nop");
            asm volatile("nop");
            asm volatile("nop");
            mreqPin_high; 
          }
        }
      }
      
      // Disable RAM
      writeByte(0x0000, 0x00);
      Serial.flush(); // Flush any serial data that wasn't processed
    }
  }
  
  rd_wr_mreq_off();
}

uint8_t readByte(int address) {
  shiftoutAddress(address); // Shift out address

  mreqPin_low;
  rdPin_low;
  asm volatile("nop"); // Delay a little (minimum is 2 nops, using 3 to be sure)
  asm volatile("nop");
  asm volatile("nop");
  uint8_t bval = ((PINB << 6) | (PIND >> 2)); // Read data
  rdPin_high;
  mreqPin_high;
  
  return bval;
}

void writeByte(int address, uint8_t data) {
  // Set pins as outputs
  DDRB |= ((1<<PB0) | (1<<PB1)); // D8 & D9
  DDRD |= ((1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7)); // D2 to D7
  
  shiftoutAddress(address); // Shift out address
  
  // Clear outputs and set them to the data variable
  PORTB &= ~((1<<PB0) | (1<<PB1)); // D8 & D9
  PORTD &= ~((1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7)); // D2 to D7
  PORTD |= (data << 2);
  PORTB |= (data >> 6);
  
  // Pulse WR
  wrPin_low;
  asm volatile("nop");
  wrPin_high;
  
  // Set pins as inputs
  DDRB &= ~((1<<PB0) | (1<<PB1)); // D8 & D9
  DDRD &= ~((1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7)); // D2 to D7
}

// Use the shift registers to shift out the address
void shiftoutAddress(unsigned int shiftAddress) {
  SPI.transfer(shiftAddress >> 8);
  SPI.transfer(shiftAddress & 0xFF);

  latchPin_low;
  asm volatile("nop");
  latchPin_high;
}

// Turn RD, WR and MREQ to high so they are deselected (reset state)
void rd_wr_mreq_reset(void) {
  rdPin_high; // RD off
  wrPin_high; // WR off
  mreqPin_high; // MREQ off
}

// Turn RD, WR and MREQ off as no power should be applied to GB Cart
void rd_wr_mreq_off(void) {
  rdPin_low;
  wrPin_low;
  mreqPin_low;
}

ISR (PCINT1_vect) {
  _delay_ms(10);
  if (PINC & (1<<PC1)) { // Invert mosfet control pin
    PORTC ^= (1<<PC0);
  }
  
  while (PINC & (1<<PC1)); // Wait for button release
  _delay_ms(250);
  
  // Check if mosfet is on or off (inverted as it's a P mosfet)
  if (PINC & (1<<PC0)) {
    SPI.end(); // End SPI
    rd_wr_mreq_off(); // Set everying low
    latchPin_low;
    dataPin_low;
    clockPin_low;
    _delay_ms(10);
    asm volatile ("jmp 0"); // Jump to address 0 (soft reset)
  }
  else {
    SPI.begin(); // Setup SPI
    SPI.setBitOrder(MSBFIRST);
    SPI.setDataMode(SPI_MODE0);
    SPI.setClockDivider(SPI_CLOCK_DIV2);
  }
} 
